from typing import List

import pytest
from koala import BaseRelationalDB

relationaldb = BaseRelationalDB


def test_relationaldbmetadata_one_cls():
    """Test the metadata of one simple class"""
    db = relationaldb()

    @db.register
    class Person:
        name: str = db.attribute()

    # Check class metadata
    db_metadata = db.db_metadata
    assert len(db_metadata) == 1
    assert "Person" in db_metadata

    # check cls metadata
    person = db_metadata["Person"]
    assert len(person) == 1
    assert "name" in person
    assert person.exist
    assert person.cls is Person

    # check attr metadata
    name_attr = person["name"]
    assert name_attr.name == "name"
    assert name_attr.type == str
    assert name_attr.atomic_type == str
    assert name_attr.atomic_type_name == "str"
    assert not name_attr.unique
    assert not name_attr.many
    assert not name_attr.ref
    assert not name_attr.referenced_by
    assert name_attr.exist


def test_relationaldbmetadata_one_cls_many_simple_attributes():
    """Test the metadata of one simple class"""
    db = relationaldb()

    @db.register
    class Person:
        name: str = db.attribute()
        nickname: str = db.attribute()
        age: int = db.attribute()
        adult: bool = db.attribute()

    # Check class metadata
    db_metadata = db.db_metadata
    assert len(db_metadata) == 1
    assert "Person" in db_metadata

    # check cls metadata
    person = db_metadata["Person"]
    assert person.exist
    assert len(person) == 4
    assert person.cls is Person

    for attribute_name in ["name", "nickname", "age", "adult"]:
        assert attribute_name in person

        # check attr metadata
        attribute = person[attribute_name]
        assert attribute.name == attribute_name
        assert not attribute.unique
        assert not attribute.many
        assert not attribute.ref
        assert not attribute.referenced_by
        assert attribute.exist

        if attribute_name in ["name", "nickname"]:
            assert attribute.type == str
            assert attribute.atomic_type_name == "str"
            assert attribute.atomic_type == str
        elif attribute_name == "age":
            assert attribute.type == int
            assert attribute.atomic_type == int
            assert attribute.atomic_type_name == "int"
        else:
            assert attribute.type == bool
            assert attribute.atomic_type == bool
            assert attribute.atomic_type_name == "bool"


def test_relationaldbmetadata_one_cls_with_cls_type():
    """Test the metadata of one simple class"""
    db = relationaldb()

    @db.register
    class Person:
        name: str = db.attribute()
        animal: "Animal" = db.attribute()

    # Check class metadata
    db_metadata = db.db_metadata
    assert len(db_metadata) == 2
    assert "Person" in db_metadata

    # check cls Person metadata
    person = db_metadata["Person"]
    assert person.exist
    assert len(person) == 2
    assert "name" in person
    assert person.cls is Person

    # check attr name metadata
    name_attr = person["name"]
    assert name_attr.name == "name"
    assert name_attr.type == str
    assert name_attr.atomic_type == str
    assert name_attr.atomic_type_name == "str"
    assert not name_attr.unique
    assert not name_attr.many
    assert not name_attr.ref
    assert not name_attr.referenced_by
    assert name_attr.exist

    # check animal attribute metadata
    animal_attr = person["animal"]
    assert animal_attr.name == "animal"
    assert animal_attr.type == "Animal"
    assert animal_attr.atomic_type == "Animal"
    assert animal_attr.atomic_type_name == "Animal"
    assert not animal_attr.unique
    assert not animal_attr.many
    assert not animal_attr.ref
    assert not animal_attr.referenced_by
    assert animal_attr.exist

    # check cls Animal metadata (at this stage we don't have information about the cls)
    animal = db_metadata["Animal"]
    assert len(animal) == 0
    assert not animal.exist


def test_relationaldbmetadata_two_simple_classes():
    """Test the metadata of one simple class"""
    db = relationaldb()

    @db.register
    class Person:
        name: str = db.attribute()

    @db.register
    class Animal:
        name: str = db.attribute()

    # Check class metadata
    db_metadata = db.db_metadata
    assert len(db_metadata) == 2
    assert "Person" in db_metadata
    assert "Animal" in db_metadata

    # check cls Person metadata
    person = db_metadata["Person"]
    assert person.exist
    assert len(person) == 1
    assert "name" in person
    assert person.cls is Person

    # check attr name metadata
    name_attr = person["name"]
    assert name_attr.name == "name"
    assert name_attr.type == str
    assert name_attr.atomic_type == str
    assert name_attr.atomic_type_name == "str"
    assert not name_attr.unique
    assert not name_attr.many
    assert not name_attr.ref
    assert not name_attr.referenced_by
    assert name_attr.exist

    # check cls Animal metadata
    animal = db_metadata["Animal"]
    assert animal.exist
    assert len(animal) == 1
    assert "name" in animal
    assert animal.cls is Animal

    # check attr name metadata
    name_attr = animal["name"]
    assert name_attr.name == "name"
    assert name_attr.type == str
    assert name_attr.atomic_type == str
    assert name_attr.atomic_type_name == "str"
    assert not name_attr.unique
    assert not name_attr.many
    assert not name_attr.ref
    assert not name_attr.referenced_by
    assert name_attr.exist


def test_relationaldbmetadata_two_classes_with_ref():
    """Test the metadata of one simple class"""
    db = relationaldb()

    @db.register
    class Person:
        name: str = db.attribute()
        animal: "Animal" = db.attribute(ref="owner")

    # ================================= #
    # Test metadata before second class #
    # ================================= #

    # Check DB metadata
    db_metadata = db.db_metadata
    assert len(db_metadata) == 2
    assert "Person" in db_metadata
    assert "Animal" in db_metadata

    # check cls Person metadata
    person = db_metadata["Person"]
    assert person.exist
    assert len(person) == 2
    assert "name" in person
    assert "animal" in person
    assert person.cls is Person

    # check attr name metadata
    name_attr = person["name"]
    assert name_attr.name == "name"
    assert name_attr.type == str
    assert name_attr.atomic_type == str
    assert name_attr.atomic_type_name == "str"
    assert not name_attr.unique
    assert not name_attr.many
    assert not name_attr.ref
    assert not name_attr.referenced_by
    assert name_attr.exist

    # check animal attribute metadata
    animal_attr = person["animal"]
    assert animal_attr.name == "animal"
    assert animal_attr.type == "Animal"
    assert animal_attr.atomic_type == "Animal"
    assert animal_attr.atomic_type_name == "Animal"
    assert not animal_attr.unique
    assert not animal_attr.many
    assert animal_attr.ref == "owner"
    assert not animal_attr.referenced_by
    assert animal_attr.exist

    # check cls Animal metadata
    animal = db_metadata["Animal"]
    assert not animal.exist
    assert len(animal) == 1
    assert "owner" in animal
    assert animal.cls is None

    # check animal attribute metadata
    owner_attr = animal["owner"]
    assert owner_attr.name == "owner"
    # assert owner_attr.type is None
    assert owner_attr.atomic_type is Person
    assert owner_attr.atomic_type_name == "Person"
    assert not owner_attr.unique
    assert not owner_attr.many
    assert not owner_attr.ref
    assert owner_attr.referenced_by is animal_attr
    assert not owner_attr.exist

    @db.register
    class Animal:
        name: str = db.attribute()
        owner: Person = db.attribute()

    assert len(db_metadata) == 2

    # change for owner_attr
    assert owner_attr.name == "owner"
    assert owner_attr.type is Person
    assert owner_attr.atomic_type is Person
    assert owner_attr.atomic_type_name == "Person"
    assert not owner_attr.unique
    assert not owner_attr.many
    assert not owner_attr.ref
    assert owner_attr.referenced_by is animal_attr
    assert owner_attr.exist


def test_relationaldbmetadata_one_cls_empty_ref():
    """Test the metadata of one simple class referencing an attribute without the second cls"""
    db = relationaldb()

    @db.register
    class Person:
        name: str = db.attribute()
        animal: "Animal" = db.attribute(ref="owner")

    # Check class metadata
    db_metadata = db.db_metadata
    assert len(db_metadata) == 2
    assert "Person" in db_metadata

    # check cls Person metadata
    person = db_metadata["Person"]
    assert person.exist
    assert len(person) == 2
    assert "name" in person
    assert person.cls is Person

    # check attr name metadata
    name_attr = person["name"]
    assert name_attr.name == "name"
    assert name_attr.type == str
    assert name_attr.atomic_type == str
    assert name_attr.atomic_type_name == "str"
    assert not name_attr.unique
    assert not name_attr.many
    assert not name_attr.ref
    assert not name_attr.referenced_by
    assert name_attr.exist

    # check attr animal metadata
    animal_attr = person["animal"]
    assert animal_attr.name == "animal"
    assert animal_attr.type == "Animal"
    assert animal_attr.atomic_type == "Animal"
    assert animal_attr.atomic_type_name == "Animal"
    assert not animal_attr.unique
    assert not animal_attr.many
    assert animal_attr.ref == "owner"
    assert not animal_attr.referenced_by
    assert animal_attr.exist

    # check cls Animal metadata (at this stage we don't have information about the cls)
    animal = db_metadata["Animal"]
    assert len(animal) == 1
    assert not animal.exist

    # Check attribute owner
    owner_attribute = animal["owner"]
    assert not owner_attribute.exist
    assert owner_attribute.name == "owner"
    # assert owner_attribute.type is None  # we can't now if it's a List or not
    assert owner_attribute.atomic_type == Person
    assert owner_attribute.atomic_type_name == "Person"
    assert not owner_attribute.unique
    assert owner_attribute.many is None
    assert owner_attribute.ref is None
    assert owner_attribute.referenced_by is animal_attr
    assert not owner_attribute.exist


# TODO: test two class
# TODO: test two classes with reference

# TODO: test if builtin type
@pytest.mark.skip
def test_relationaldbmetadata_one_cls_relationaldb_metadata():
    """Test the metadata of one simple class"""
    db = relationaldb()

    @db.register
    class Person:
        name: str = db.attribute(unique=True)
        animal: "Animal" = db.attribute(ref="owner")

    # Check class metadata
    db_metadata = db.db_metadata
    assert len(db_metadata) == 2
    assert "Person" in db_metadata
    assert "Animal" in db_metadata

    # check cls metadata
    person = db_metadata["Person"]
    assert len(person) == 2
    assert person.cls is Person

    # check name metadata
    # check attr metadata
    assert "name" in person
    name_attr = person["name"]
    assert name_attr.name == "name"
    assert name_attr.type == str
    assert name_attr.atomic_type_name == "str"
    assert not name_attr.unique
    assert not name_attr.many
    assert name_attr.ref == "owner"
    assert not name_attr.referenced_by


def test_error_ref_builtin_attribute():
    for type in [str, int, float, bool]:
        db = relationaldb()

        with pytest.raises(ValueError):
            # We cannot reference built-in attribute
            @db.register
            class Person:
                attribute: type = db.attribute(ref="attr")


def test_error_ref_no_existing_attribute():
    db = relationaldb()

    @db.register
    class Person:
        name: db.attribute()

    with pytest.raises(ValueError):

        @db.register
        class Animal:
            name: db.attribute()
            owner: Person = db.attribute(ref="animal")


def test_relationaldb_metadata_one_cls_with_list():
    db = relationaldb()

    @db.register
    class Person:
        name: str = db.attribute()
        animals: List["Animal"] = db.attribute()

    # Check class metadata
    db_metadata = db.db_metadata
    assert len(db_metadata) == 1
    person = db_metadata["Person"]

    assert len(person) == 2
    assert "animals" in person

    # check animals metadata
    animals_attribute = person["animals"]
    assert animals_attribute.name == "animals"
    assert animals_attribute.type == List["Animal"]
    assert animals_attribute.atomic_type_name == "Animal"
    assert not animals_attribute.unique
    assert animals_attribute.many
    assert not animals_attribute.ref
    assert not animals_attribute.referenced_by


@pytest.mark.skip(reason="Not implemented yet")
def test_many_many():
    db = relationaldb()

    @db.register
    class Person:
        name: str = db.attribute(unique=True)
        animals: List["Animal"]

    @db.register
    class Animal:
        name: str
        owners: List[Person] = db.attribute(ref="animals")

    foo = Person("foo")
    bar = Person("bar")
    ploofy = Animal("ploofy")
    rex = Animal("rex")

    foo.animals.append(ploofy)
    assert ploofy in foo.animals
    assert foo in ploofy.owners

    db.first("Animal", name="foo")



# ================
import datetime
import random
import tempfile

import attr
import pytest
from sqlalchemy import (
    Boolean,
    Column,
    Date,
    DateTime,
    ForeignKey,
    Integer,
    String,
    create_engine,
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker

engine = create_engine("sqlite:///:memory:", echo=False)
Session = sessionmaker(bind=engine)
Base = declarative_base()

"""
for a simple one-to-many there are 2 differents interpratation for relationship
Ex: class Person:  best_friend_id  ...
- the object pointing to (the best_friend)
- all object pointing to me (i am the bestfriend of...) (BY DEFAULT)
# by default p


Ex2:   class Animal: owner_id
- Animal: owner
- Person: animals
"""


class Person(Base):
    __tablename__ = "persons"
    id = Column(Integer, primary_key=True)
    name: str = Column(String)
    age: int = Column(Integer)
    best_friend_id = Column(Integer, ForeignKey("persons.id"))
    ref = relationship("Person")
    # best_friend = relationship("Person", foreign_keys=best_friend_id, uselist=False)
    # best_friend_of = relationship("Person", remote_side=id)
    # best_friend = relationship("Person")

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"


Base.metadata.create_all(engine)
session = Session()


foo = Person(name="foo", age=20)
bar = Person(name="bar", age=21)
print(foo.red)


engine = create_engine("sqlite:///:memory:", echo=False)
Session = sessionmaker(bind=engine)
Base = declarative_base()


class Customer(Base):
    __tablename__ = "customer"
    id = Column(Integer, primary_key=True)
    name = Column(String)

    billing_address_id = Column(Integer, ForeignKey("address.id"))
    shipping_address_id = Column(Integer, ForeignKey("address.id"))
    customer_id = Column(Integer, ForeignKey("customer.id"))

    billing_address = relationship("Address", foreign_keys=billing_address_id)
    customer = relationship("Customer", foreign_keys=customer_id)
    shipping_address = relationship("Address", foreign_keys=[shipping_address_id])

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"


class Address(Base):
    __tablename__ = "address"
    id = Column(Integer, primary_key=True)
    street = Column(String)
    city = Column(String)
    state = Column(String)
    zip = Column(String)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.city})"


Base.metadata.create_all(engine)
session = Session()

a1 = Address(street="12", city="city", state="state", zip="zip")
a2 = Address(street="13", city="2", state="2", zip="2")

session.add(a1)
session.add(a2)
session.commit()
foo = Customer(name="foo")
bar = Customer(name="bar")
print(foo.customer)


# ==================
engine = create_engine("sqlite:///:memory:", echo=False)
Session = sessionmaker(bind=engine)
Base = declarative_base()


class Person(Base):
    __tablename__ = "persons"
    id = Column(Integer, primary_key=True)
    name: str = Column(String)
    age: int = Column(Integer)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"


class Animal(Base):
    __tablename__ = "animals"
    id = Column(Integer, primary_key=True)
    owner_id = Column(Integer, ForeignKey("persons.id"))
    name: str = Column(String)
    owner = relationship("Person", foreign_keys="Animal.owner_id")

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"


Base.metadata.create_all(engine)
session = Session()

foo = Person(name="foo", age=20)
bar = Person(name="bar", age=21)

session.add(foo)
session.add(bar)
session.commit()


afoo = Animal(name="afoo", owner=foo)
session.add(afoo)
session.commit()

# STRATEGY: apply attr => inherit from attr

# ==================== TEST BUDGET ============
engine = create_engine("sqlite:///:memory:", echo=False)
Session = sessionmaker(bind=engine)
Base = declarative_base()


@attr.s
class BudgetBase:
    name: str = attr.ib()
    initial_balance: int = attr.ib(default=0)
    description: str = attr.ib(default=True, kw_only=True)
    creation_date = attr.ib(factory=datetime.datetime.utcnow)
    loan: bool = attr.ib(default=True, kw_only=True)
    frm = attr.ib(default=None, kw_only=True)
    balance = attr.ib(init=False)
    refunded_balance = attr.ib(init=False)

    def __attrs_post_init__(self):
        self.balance = self.initial_balance
        self.refunded_balance = self.initial_balance


old_init = BudgetBase.__attrs_post_init__


def new_init(self):
    old_init(self)
    session.add(self)
    session.commit()


BudgetBase.__attrs_post_init__ = new_init


class Budget(BudgetBase, Base):
    __tablename__ = "budgets"

    id = Column(Integer, primary_key=True)
    name = Column(String)
    description = Column(String)
    creation_date = Column(DateTime, default=datetime.datetime.utcnow, nullable=False)
    loan = Column(Boolean, default=True, nullable=False)
    id_from = Column(Integer, ForeignKey("budgets.id"), nullable=True)

    # balance
    initial_balance = Column(Integer, nullable=False)
    balance = Column(Integer, nullable=False)
    refunded_balance = Column(Integer, nullable=False)

    @property
    def transactions(self):
        return session.query(Transaction).filter_by(id_budget=self.id)


@attr.s
class TransactionBase:
    budget: Budget = attr.ib()
    value: int = attr.ib()
    description: str = attr.ib(default="")
    affected: Budget = attr.ib(default=None)
    loan: bool = attr.ib(default=False)

    def __attrs_post_init__(self):
        pass


old_init2 = TransactionBase.__attrs_post_init__


def new_init(self):
    old_init2(self)
    session.add(self)
    session.commit()


BudgetBase.__attrs_post_init__ = new_init


class Transaction(TransactionBase, Base):
    __tablename__ = "transactions"
    id = Column(Integer, primary_key=True)
    id_budget = Column(Integer, ForeignKey("budgets.id"), nullable=False)
    id_affected_budget = Column(Integer, ForeignKey("budgets.id"), nullable=True)
    time = Column(DateTime, default=datetime.datetime.utcnow, nullable=False)
    value = Column(Integer, nullable=False)
    description = Column(String, nullable=False)
    loan = Column(Boolean, default=True, nullable=False)

    # transactions
    @property
    def budget(self):
        return session.query(Budget).filter_by(id=self.id_budget).first()

    @budget.setter
    def budget(self, value):
        self.id_budget = value.id


bnp = Budget("bnp")
print(bnp)

t1 = Transaction(bnp, 100, "phone")
print(t1)

# =================== END TEST ================
attributes = {
    "id": Column(Integer, primary_key=True),
    "name": Column(String),
    "age": Column(Integer),
}
x = Person.__attrs_attrs__


class Budget(Base):
    __tablename__ = "budgets"

    id = Column(Integer, primary_key=True)
    name = Column(String)
    description = Column(String)
    creation_date = Column(DateTime, default=datetime.datetime.utcnow, nullable=False)
    loan = Column(Boolean, default=True, nullable=False)
    id_from = Column(Integer, ForeignKey("budgets.id"), nullable=True)

    # balance
    initial_balance = Column(Integer, nullable=False)
    balance = Column(Integer, nullable=False)
    refunded_balance = Column(Integer, nullable=False)

    def __init__(
        self,
        name,
        initial_balance,
        description="",
        creation_date=None,
        loan=False,
        frm=None,
    ):
        self.name = name
        self.initial_balance = initial_balance
        self.description = description
        if creation_date is None:
            self.creation_date = datetime.datetime.utcnow()
        self.loan = loan
        self.frm = frm
        self.balance = initial_balance
        self.refunded_balance = initial_balance
        session.add(self)
        session.commit()

    def __repr__(self):
        return f"Budget({self.name})"

    @property
    def transactions(self):
        return session.query(Transaction).filter_by(id_budget=self.id)


class Transaction(Base):
    __tablename__ = "transactions"
    id = Column(Integer, primary_key=True)
    id_budget = Column(Integer, ForeignKey("budgets.id"), nullable=False)
    id_affected_budget = Column(Integer, ForeignKey("budgets.id"), nullable=True)
    time = Column(DateTime, default=datetime.datetime.utcnow, nullable=False)
    value = Column(Integer, nullable=False)
    description = Column(String, nullable=False)
    loan = Column(Boolean, default=True, nullable=False)

    def __init__(self, budget, value, description="", affected=None, loan=False):
        self.budget = budget
        self.value = value
        self.description = description
        self.affected = affected
        self.loan = loan
        session.add(self)
        session.commit()

    def __repr__(self):
        return f"Transaction({self.description}, {self.value})"

    # transactions
    @property
    def budget(self):
        return session.query(Budget).filter_by(id=self.id_budget).first()

    @budget.setter
    def budget(self, value):
        self.id_budget = value.id


Base.metadata.create_all(engine)
session = Session()


bnp = Budget("bnp", 1000)
mama = Budget("mama", 2000)
print(bnp)
print(mama)

t1 = Transaction(bnp, 100, "bazef")
t2 = Transaction(bnp, 50, "test")
for i in range(20):
    print(i)
    budget = random.choice([bnp, mama])
    value = random.randint(-100, 100)
    des = random.choice(["phone", "test", "some", "thing", "well"])
    print(budget, value, des)
    Transaction(budget, value, des)


print(t1)
print(t2)

print(bnp.transactions)
for t in bnp.transactions:
    print(t)



import datetime
import tempfile

import attr
import pytest
from sqlalchemy import (
    Boolean,
    Column,
    Date,
    DateTime,
    ForeignKey,
    Integer,
    String,
    create_engine,
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker


@attr.s
class Budget:
    name = attr.ib()
    initial_balance = attr.ib(default=0)
    transactions = attr.ib(factory=list)
    frm = attr.ib(default=None)
    loan = attr.ib(default=False)
    balance = attr.ib(init=False)

    def __attrs_post_init__(self):
        self.balance = self.initial_balance


@attr.s
class Transaction:
    # budget = attr.ib(repr=False)
    value = attr.ib()
    description = attr.ib()
    # If the transaction affect an other budget
    affect = attr.ib(default=None)
    loan = attr.ib(default=False)


engine = create_engine("sqlite:///:memory:", echo=True)
Session = sessionmaker(bind=engine)
Base = declarative_base()


attributes = {
    "__tablename__": "budgets",
    "id": Column(Integer, primary_key=True),
    "name": Column(String),
    "loan": Column(Boolean, default=True, nullable=False),
    "frm": Column(Integer, ForeignKey("budgets.id"), nullable=True),
    "initial_balance": Column(Integer, default=0, nullable=False),
    "balance": Column(Integer, nullable=False),
    # "transactions": property()
    "__init__": Budget.__init__,
    "__attrs_post_init__": Budget.__attrs_post_init__,
    "__repr__": Budget.__repr__,
}
BudgetORM = type("BudgetORM", (Base,), attributes)

# transactions
@property
def budget(self):
    session.query(BudgetORM).filter_by(id=self.id_budget).first()


@budget.setter
def budget(self, value):
    self.id_budget = value.id


attributes = {
    "__tablename__": "transactions",
    "id": Column(Integer, primary_key=True),
    "id_budget": Column(Integer, ForeignKey("budgets.id"), nullable=False),
    "id_affected_budget": Column(Integer, ForeignKey("budgets.id"), nullable=True),
    "value": Column(Integer, default=0, nullable=False),
    "description": Column(Integer, nullable=False),
    "loan": Column(Boolean, default=True, nullable=False),
    # "transactions": property()
    "__init__": Transaction.__init__,
    "__repr__": Transaction.__repr__,
    "budget": budget,
}

TransactionORM = type("TransactionORM", (Base,), attributes)

Base.metadata.create_all(engine)
session = Session()

bnp = BudgetORM("bnp", 1000)
print(bnp)
session.add(bnp)
session.commit()

t = TransactionORM(bnp, 100, "bazef")
print(t)



import io
import os
import random
import shutil
import string
import sys
import tempfile
import types
from contextlib import contextmanager
from typing import Any, Dict, Iterable


@contextmanager
def tmp_module(
    module_name: str = None,
    context: Dict[str, Any] = None,
    addsys: bool = True,
    overwrite=False,
):
    """Generate a temporal module as context.
    Temporally add the module to sys.modules so it can be importable
    Args:
    """
    # TODO: unit test this function
    if not module_name:
        module_name = "tmpmodule_" + randomstr(exclude=list(sys.modules.keys()))
    else:
        if module_name in sys.modules and not overwrite:
            raise TabError("Module name already exist, call it with overwrite = True")

    context = context or {}
    module = types.ModuleType(module_name)
    module.__dict__.update(context)
    sys.modules[module_name] = module
    yield module
    del sys.modules[module_name]


@contextmanager
def temp_pathname():
    """Return an available pathname that you can use for files/dirs

    Will automatically clean the pathname if it's
    a file or directory"""
    tmpdir = tempfile.gettempdir()
    pathname = os.path.join(
        tmpdir, randomstr(20, prefix="tmppathname_", exclude=os.listdir(tmpdir))
    )
    try:
        yield pathname
    finally:
        if os.path.isdir(pathname):
            shutil.rmtree(pathname)
        elif os.path.isfile(pathname):
            os.remove(pathname)


def randomstr(
    n: int = 10,
    *,
    prefix="",
    alphabet: str = string.ascii_letters,
    exclude: Iterable = None
):
    """Generate random string of length n
    Args:
        n: length of returned string (without the prefix)
        alphabet: alphabet to use
        exclude: generated string must not be in this list (or any iterable object)
    """
    if exclude is not None:
        exclude = list(exclude)
    else:
        exclude = []

    while True:
        string = prefix + "".join(random.choices(alphabet, k=n))
        if string not in exclude:
            return string


@contextmanager
def temp_str2filename(string: str) -> str:
    """Generate a temporary file with the string content and
    return it's filename
    """
    with temp_pathname() as pathname:
        with open(pathname, "w") as f:
            f.write(string)
        yield pathname


@contextmanager
def str2stdin(string: str):
    string_io = io.StringIO(string)
    old_stdin = sys.stdin
    sys.stdin = string_io
    yield
    sys.stdin = old_stdin


@contextmanager
def file2stdin(filename: str):
    with open(filename) as f:
        old_stdin = sys.stdin
        sys.stdin = f
        try:
            yield
        # FIXME unit test finally with all other contextmanager
        finally:
            sys.stdin = old_stdin



import datetime
import tempfile

import attr
import pytest
from sqlalchemy import (
    Boolean,
    Column,
    Date,
    DateTime,
    ForeignKey,
    Integer,
    String,
    create_engine,
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker


# ===== #
# TESTS #
# ===== #
def tes_relational_one_to_n():
    class Person:
        def __init__(self, name):
            self.name = name
            self.animals = []

    class Animal:
        def __init__(self, owner, name):
            self.owner = owner
            self.name = name

    person = Person("person")
    animal = Animal(person, "animal")


def t2():
    @attr.s
    class Person:
        name = attr.ib()
        # 0:n
        animals = relation("0:n")

    @attr.s
    class Animal:
        owner = relation("0:1", Person.animals)
        name = attr.ib()

    person = Person("person")
    animal = Animal(person, "animal")
    assert animal in person.animals

    animal2 = Animal("woofy")
    person2 = Person("person2")
    person2.animals.append(animal2)
    assert animal2.owner is person2


def t_db_0_n_to_0_n():
    class Person:
        name = attribute()
        animals = attribute("n")

    class Animal:
        name = attribute()
        owners = attribute("n", Person.animals)

    p1 = Person("p1")
    a1 = Animal("a1")
    p1.animals.append(a1)
    assert p1 in a1.owners


def t_db_one_to_many():
    db = Relational("budgetm")

    @db.table("transactions")
    @attr.s
    class Transaction:
        value = attr.ib()

    @db.table("budgets")
    @attr.s
    class Budget:
        name = attr.ib()
        transactions = attr.ib(factory=list)

    bnp = Budget("bnp")
    bnp.transactions.append(Transaction(30))
    bnp.transactions.append(Transaction(-20))
    print(bnp)


def t272414():
    db = Relational()

    @db.table("budgets")  # primary key is id
    class Budget:
        name = db.Collumn(db.String)
        transactions = db.relational(table="transactions", target="budget")
        frm_budget = db.Collumn("budgets", nullable=True)

    @db.table("transactions")
    class Transaction:
        value = db.Collumn(db.Integer)
        budget = ...


def tes_relationaldb_simple():
    @attr.s
    class Animal:
        name: str = attr.ib()
        age: int = attr.ib()

    with tempfile.NamedTemporaryFile() as f:
        db = Relational(f.name, echo=True)
        db.table("animals")(Animal)
        db.init()

        animal = Animal("woofy", 25)
        print(animal)
        db = Relational(f.name, echo=True)
        db.table("animals")(Animal)
        db.init()

        assert db.query(Animal).first(name="woofy").age == 25


class Test:
    name: str = attr.ib()
    age: int = attr.ib()
    age2: int = attr.ib(init=False)

    def __attrs_post_init__(self):
        self.age2 = self.age


def test_relationaldb_relation():
    db = Relational(echo=True)

    @db.table("budgets")
    class Budget:
        name: str = db.attribute(unique=True)
        initial_balance: int = db.attribute()
        # from now only accept key words only
        # kw_only = db.kw_only()# TODO: later
        description: str = db.attribute(default="", kw_only=True)
        frm = db.attribute(type="budgets", default=None, kw_only=True)
        creation_date: DateTime = db.attribute(factory=db.datetime_now, kw_only=True)
        loan: bool = db.attribute(default=False, kw_only=True)

        # balance
        balance: int = db.attribute(init=False)
        refunded_balance: int = db.attribute(init=False)

        def __attrs_post_init__(self):
            self.balance = self.initial_balance
            self.refunded_balance = self.initial_balance

    @db.table("transactions")
    class Transaction:
        budget: Budget = db.attribute(target="transactions")
        value: int = db.attribute()
        description: str = db.attribute()
        # _kw_only = db.kw_only()
        affected_budget = db.attribute(type="budgets", default=None, kw_only=True)
        time: DateTime = db.attribute(factory=db.datetime_now, kw_only=True)
        loan: bool = db.attribute(default=False, kw_only=True)

    db.init()

    bnp = Budget("bnp", 1000)
    bnp_fromdb = db.query(Budget).filter_by(name="bnp").first()
    assert bnp is bnp_fromdb

    t1 = Transaction(bnp, 300, "printer")
    assert t1 in bnp.transactions
    assert list(bnp.transactions) == [t1]

    assert t1.budget is bnp

    # get bnp budget


def tes_relationaldb_select():
    db = Relational(echo=True)

    @db.table("animals")
    class Animal:
        name: str = db.attribute()
        age: int = db.attribute()
        strong: bool = db.attribute(default=False, kw_only=True)

    db.init()

    # session = db.session

    a1 = Animal("woofy", 12)
    assert a1.age == 12
    assert a1.name == "woofy"

    a2 = Animal("doggy", 10)
    a3 = Animal("Garry", 8)

    # Get animals that have an age greater or equal than 10
    # session.query(Animal).filter(Animal.a)
    animals = [db.query(Animal.orm).filter(Animal.orm.age >= 10)]
    assert animals[0] is a1
    assert animals[1] is a2
    assert len(animals) == 2


def x():
    @attr.s
    class Budget:
        name: String = attr.ib()
        caca: int = attr.ib(default="haha")

    bnp = Budget("bnp", "lol")
    print(bnp)


def test_relationaldb():
    with tempfile.NamedTemporaryFile() as f:
        db = Relational(f.name)

        @db.table("persons")
        class Person:
            name: str = db.attribute()
            age: int = db.attribute()

        db.init()

        Person("nazime", 25)
        Person("massil", 28)

        db = Relational(f.name)

        @db.table("persons")
        class Person:
            name: str = db.attribute()
            age: int = db.attribute()

        db.init()

        nazime = db.query(Person).filter_by(name="nazim").first()
        assert nazime.name == "nazime"
        assert nazime.age == 25



import glob
import inspect
import os
import shutil
from string import Template
from typing import Any, Callable, Dict, List, Union

from koalak.frameworks import home_structure
from koalak.frameworks.home_structure import (
    DirectoryDescription,
    FileDescription,
    normalize_home_structure,
)

from . import databases, exceptions
from .consts import HOMEPATH_NAME, PACKAGE_NAME


def __init_subclass__(subcls, **kwargs):

    if "name" not in subcls.__dict__:
        raise ValueError(f"Base plugin must have the attribute 'name'")

    # add the plugin
    name = subcls.__dict__["name"]
    pm = subcls._fwm["plugin_manager"]
    if name in pm._plugins:
        raise exceptions.PluginAlreadyExistException(f"plugin {name} already exist")

    pm._plugins[name] = subcls

    # abstract class
    if FwmConfig.is_abstract(subcls):
        pass


def init_subclass(subcls, **kwargs):
    subcls._pluginmanager._register(subcls)


class FwmConfig:
    """Class responsible to handle internal config in objects
    like plugins. Ex: @abstract"""

    dict_name = PACKAGE_NAME

    @classmethod
    def update(cls, key, value):
        pass

    @classmethod
    def get(cls, key):
        pass

    @classmethod
    def set_abstract(cls, target_cls):
        cls.update("abstract", True)

    @classmethod
    def is_abstract(cls, target_cls):
        pass


class Plugin:
    pass


def mkpluginmanager(*args, **kwargs):
    return PluginManager(*args, **kwargs)


class ConstraintAttr:
    def __init__(self, name: str = None, type=None, inheritable: bool = True):
        """

        Args:
            name: name of the class attribute
            type: type of the class attribute
            inheritable: if the attribute can be inherited (True)
                or if each class must define the class attribute (False)
        """
        self.name = name
        self.type = type
        self.inheritable = inheritable

    def check(self, plugin):
        if self.name is None:
            raise TypeError("Must set the attribute name")
        if self.inheritable:
            has_attr = hasattr(plugin, self.name)
        else:
            has_attr = self.name in plugin.__dict__

        if not has_attr:
            raise TypeError(
                f"plugin {plugin.__name__} must have the class attribute {self.name}"
            )

        plugin_attr = getattr(plugin, self.name)
        # Check type
        if self.type is not None:
            if not isinstance(plugin_attr, self.type):
                raise TypeError(
                    f"class attribute {self.name} must be of type {self.type}"
                )


class PluginManager:
    """

    How it work:

    """

    def __init__(self, name: str = None, *, homepath=None):
        # FIXME: plugin without name?
        self.name: str = name
        self.baseplugin = None
        self.homepath = homepath
        self._plugins = {}
        self._initialized = False
        self._constraints = []

    def __repr__(self):
        if self.name:
            return f"<PluginManager [{self.name}]>"
        else:
            return f"<PluginManager>"

    def __str__(self):
        return self.__repr__()

    def _assert_baseplugin(self):
        if self.baseplugin is None:
            raise TabError("BasePlugin not inited!")

    def __getitem__(self, item: str):
        return self._plugins[item]

    def __iter__(self):
        yield from self._plugins.values()

    def _register(self, plugin):
        if self.baseplugin is None:
            raise TypeError(
                "Can't register plugin without baseplugin. use mkbaseplugin before"
            )

        # Check constraint before registring
        for contraint in self._constraints:
            contraint.check(plugin)

        self._plugins[plugin.name] = plugin

    def mkbaseplugin(self, baseplugin):
        self.baseplugin = baseplugin
        baseplugin._pluginmanager = self
        baseplugin.__init_subclass__ = classmethod(init_subclass)

        attrs_to_del = []  # we can't del in loop
        for name, constraint in baseplugin.__dict__.items():
            if isinstance(constraint, ConstraintAttr):
                constraint.name = name
                self._constraints.append(constraint)
                attrs_to_del.append(name)
        for attr_name in attrs_to_del:
            delattr(baseplugin, attr_name)

        return baseplugin

    def init(self, _homepath_initialized: set = None):
        _homepath_initialized = _homepath_initialized or set()
        if self.homepath is None:
            raise TypeError("You can not init a plugin mananger without homepath")

        if self._initialized:
            raise TypeError("Plugin Already initiated")
        self._initialized = True

        if self.homepath in _homepath_initialized:
            return
        self._init_home()
        self._load_plugins()

    def attr(self, type=None, inheritable=True):
        return ConstraintAttr(type=type, inheritable=inheritable)

    def _load_plugins(self):
        for python_path in glob.glob(os.path.join(self.homepath, "*.py")):
            with open(python_path) as f:
                data = f.read()
                exec(data, {})

    def _init_home(self):
        if not self.homepath:
            return
        if not os.path.exists(self.homepath):
            os.makedirs(self.homepath)
        elif os.path.isfile(self.homepath):
            raise NotADirectoryError("Home path is not a directory")
        # else it's a directory already created


class HookManager:
    def __init__(self, name, spec_func):
        self.name = name
        self.spec_func = spec_func
        self._hooks = {}

    def register(self, func):
        func_name = func.__name__
        if func_name in self._hooks:
            raise exceptions.HookAlreadyExistException(
                f"hook {func_name} already exist"
            )

        self._hooks[func_name] = func
        return func

    def get_hook(self, hook_name: str) -> Callable:
        return self._hooks[hook_name]

    def get_hooks(self) -> List[Callable]:
        return list(self._hooks.values())

    def runall(self, *args, **kwargs):
        results = []
        for func in self._hooks.values():
            result = func(*args, **kwargs)
            results.append(result)
        return results


class Printer:
    pass


class Framework:
    def __init__(
        self,
        name: str = None,
        *,
        homepath: str = None,
        home_structure=None,
        variables: Dict[str, Any] = None,
        version: str = None,
    ):
        self.name: str = name
        self.homepath: str = homepath
        if version is None:
            version = "0.0.0"
        self.version: str = version
        # The framework can be initialized if it has a name or a home path
        self._can_be_initialized = name or homepath
        if home_structure is None:
            home_structure = []
        else:
            home_structure = normalize_home_structure(home_structure)
        self._home_structure = home_structure
        self._plugin_managers: Dict[str, PluginManager] = {}
        self._hook_managers: Dict[str, HookManager] = {}
        self._initialized: bool = False
        self._variables: Dict[str, Any] = {}
        # add home variable
        self.set_variable("home", self.homepath)
        self.set_variable("userhome", os.path.expanduser("~"))
        if variables:
            # FIXME: ...
            substitute = self._variables
            for key, value in variables.items():
                value = Template(value).substitute(substitute)
                substitute[key] = value
            self._variables = substitute

        self.printer = Printer()

        self._databases = {}

        if self._can_be_initialized and self.homepath is None:
            self.homepath = os.path.expanduser(
                os.path.join("~", f".{HOMEPATH_NAME}", self.name)
            )

    def __repr__(self):
        if self.name:
            return f"<Framework [{self.name}]>"
        else:
            return f"<Framework>"

    def __str__(self):
        return self.__repr__()

    def fixture(self, function):
        pass

    def init(self):
        # We can init a framework if he has a name or a homepath
        if self._initialized:
            raise TypeError(f"Framework {self.name} already initialized")
        self._initialized = True

        # Check if it can be initialized
        if not self._can_be_initialized:
            raise TypeError(
                f"Framework need a name or a homepath in order to be initialized"
            )

        # Check if it's already initialized (in previous run)
        if not self._is_initialized():
            self._init_home()

        # Init plugins
        homepath_initialized = set()  # Don't initialize twice the same homepath!
        for plugin_manager in self._plugin_managers.values():
            homepath = plugin_manager.homepath
            plugin_manager.init(_homepath_initialized=homepath_initialized)
            homepath_initialized.add(homepath)

        # TODO: at the end init the framework database (version)

    def _is_initialized(self):
        return os.path.isdir(self.homepath)

    def _init_home(self, parent_path=None, nodes=None):
        """Recursive function"""
        # FIXME: add support for nodes as list of string and dict
        # FIXME: see _normalize_home_node and skipped tests
        # TODO: add support for tree ``tree -Jd``
        # TODO: add this as a standalone function
        if parent_path is None:
            os.makedirs(self.homepath)
            parent_path = self.homepath

        if nodes is None:
            nodes = self._home_structure

        for node in nodes:
            path = os.path.join(parent_path, node.name)
            if isinstance(node, FileDescription):
                # run before_normalize hooks actions
                print("actions", node.actions)
                for action in node.actions:
                    print("action", action)
                    action.before_normalize(node, self)
                    print("after action")

                norimalized_file = home_structure.normalize_file_description(
                    node, parent_path, self
                )

                # run after_normalize hooks actions
                for action in node.actions:
                    action.after_normalize(norimalized_file)

                with open(norimalized_file.path, "w") as f:
                    f.write(norimalized_file.content)

                # run after_normalize hooks actions
                for action in node.actions:
                    action.after_creation(norimalized_file)

            elif isinstance(node, DirectoryDescription):
                os.mkdir(path)
                self._init_home(path, node.nodes)

    # @defaultargs(firstarg=str)
    def mkhkmanager(self, func, hook_name):
        pass

    def mkhookmanager(self, hook_name_or_func: Union[Callable, str]):
        def _mkhookmanager(spec_function):
            if hook_name in self._hook_managers:
                raise exceptions.HookManagerAlreadyExistException(
                    f"hook {hook_name} already exist in {self.name} framework"
                )

            hm = HookManager(hook_name, spec_function)
            self._hook_managers[hook_name] = hm
            return hm

        if isinstance(hook_name_or_func, str):
            hook_name = hook_name_or_func
        elif inspect.isfunction(hook_name_or_func):
            hook_name = hook_name_or_func.__name__
            return _mkhookmanager(hook_name_or_func)
        else:
            pass
            # TODO: raise not handled type

        return _mkhookmanager

    def mkpluginmanager(self, pluginmanager_name: str = None, homepath=None):

        if homepath is None and self.homepath:
            homepath = os.path.join(self.homepath, "plugins")
        elif homepath and self.homepath:
            homepath = os.path.join(self.homepath, homepath)

        pluginmanager = PluginManager(pluginmanager_name, homepath=homepath)
        self._plugin_managers[pluginmanager_name] = pluginmanager
        return pluginmanager

    def mkpluginmanager2(self, plugin_name: str):
        def _mkplugin(base_cls):
            if plugin_name in self._plugin_managers:
                raise exceptions.PluginManagerAlreadyExistException(
                    f"plugin manager {plugin_name} already exist"
                )
            pm = PluginManager(plugin_name, base_cls)
            self._plugin_managers[plugin_name] = pm

            base_cls._fwm = {"plugin_manager": pm}
            base_cls.__init_subclass__ = classmethod(__init_subclass__)
            return base_cls

        return _mkplugin

    def get_plugins(self, plugin_manager_name: str):
        return self._plugin_managers[plugin_manager_name].get_plugins()

    def get_plugin_manager(self, plugin_manager_name: str) -> PluginManager:
        return self._plugin_managers[plugin_manager_name]

    def get_hook_manager(self, hook_manager_name: str) -> HookManager:
        return self._hook_managers[hook_manager_name]

    def get_hook_managers(self):
        return list(self._hook_managers.values())

    def get_hook(self, hook_manager_name, hook_name):
        return self._hook_managers[hook_manager_name].get_hook(hook_name)

    def get_hooks(self, hook_manager_name):
        return self._hook_managers[hook_manager_name].get_hooks()

    def get_plugin(self, plugin_manager_name: str, plugin_name: str):
        return self._plugin_managers[plugin_manager_name].get_plugin(plugin_name)

    def create_variable(self, variable_name: str, variable_value: Any) -> None:
        if variable_name in self._variables:
            raise exceptions.VariableAlreadyExistException(
                f"variable {variable_name} already exist"
            )
        self._variables[variable_name] = variable_value

    def get_variable(self, variable_name: str) -> Any:
        return self._variables[variable_name]

    def set_variable(self, variable_name: str, variable_value: Any) -> None:
        self._variables[variable_name] = variable_value

    def get_variables(self) -> Dict:
        # FIXME: add variables of parent
        return self._variables

    def create_list_database(
        self, name, *, path=None, type=None, unique=None
    ) -> databases.ListDatabase:
        # FIXME: init before or create database? or not important?
        if path is None:
            path = "$home/.databases/name"
        path = Template(path).substitute(self.get_variables())
        db = databases.ListTxtDatabase(path, unique=unique)
        self._databases[name] = db
        return db

    def get_database(self, name):
        return self._databases[name]

    def uninstall(self):
        """Undo every thing init did"""
        # TODO: can't be called without init being called
        # TODO: test! test plugins (they might have their own path)
        # Remove the home path
        shutil.rmtree(self.homepath)

        # undo file actions
        # FIXME: must be recursive! implement iternodes
        for node in self._home_structure:
            if isinstance(node, FileDescription):
                for action in node.actions:
                    action.uninstall(node, self)

    def substitute_string(self, string: str):
        return Template(string).substitute(self.get_variables())


class FrameworkManager:
    _frameworks = {}
    _unique_framework_name_counter = 0

    @classmethod
    def mkframework(
        cls,
        framework_name: str = None,
        *,
        homepath=None,
        home_structure=None,
        variables=None,
        version: str = None,
    ) -> Framework:
        if framework_name and framework_name in cls._frameworks:
            raise exceptions.FrameworkAlreadyExistException(
                f"Framework {framework_name} already exist"
            )

        framework = Framework(
            framework_name,
            homepath=homepath,
            home_structure=home_structure,
            variables=variables,
            version=version,
        )
        cls._frameworks[framework_name] = framework
        return framework

    @classmethod
    def get_framework(cls, framework_name: str) -> Framework:
        return cls._frameworks[framework_name]

    @classmethod
    def get_frameworks(cls) -> List[Framework]:
        return list(cls._frameworks.values())

    @classmethod
    def get_unique_framework_name(cls) -> str:
        while True:
            counter = cls._unique_framework_name_counter
            name = f"unique_framework_name_{counter}"
            if name not in cls._frameworks:
                return name
            cls._unique_framework_name_counter += 1



import os

import koalak
import pytest
from koalak.utils import temp_pathname


def test_simple():
    """Test framework can be created"""
    framework = koalak.mkframework()


# @pytest.mark.skip(reason="tmp_pathname not implemented")
def test_double_init():
    """If we call init twice raise Error"""
    with temp_pathname() as pathname:
        framework = koalak.mkframework(homepath=pathname)
        framework.init()
        with pytest.raises(TypeError):
            framework.init()


# @pytest.mark.skip(reason="tmp_pathname not implemented")
def test_homepath_alone():
    """Homepath is created after init"""
    with temp_pathname() as pathname:
        framework = koalak.mkframework(homepath=pathname)
        assert not os.path.exists(pathname)
        framework.init()
        assert os.path.isdir(pathname)


def test__repr__and__str__():
    framework = koalak.mkframework()
    assert repr(framework) == str(framework) == "<Framework>"

    framework = koalak.mkframework("wordlistools")
    assert repr(framework) == str(framework) == "<Framework [wordlistools]>"


# @pytest.mark.skip(reason="tmp_pathname not implemented")
def test_homepath_plugins_alone():
    """Test that folder plugins is created"""
    with temp_pathname() as pathname:
        framework = koalak.mkframework(homepath=pathname)
        plugins = framework.mkpluginmanager()
        assert not os.path.exists(pathname)
        framework.init()
        assert os.path.isdir(pathname)
        assert os.path.isdir(os.path.join(pathname, "plugins"))


# @pytest.mark.skip(reason="tmp_pathname not implemented")
def test_homepath_plugins_with_homepath():
    """Test that we can chose the folder name of plugins"""
    with temp_pathname() as pathname:
        framework = koalak.mkframework(homepath=pathname)
        plugins = framework.mkpluginmanager(homepath="tools")
        assert not os.path.exists(pathname)
        framework.init()
        assert os.path.isdir(pathname)
        assert not os.path.isdir(os.path.join(pathname, "plugins"))
        assert os.path.isdir(os.path.join(pathname, "tools"))



import koalak
import pytest


def test_create_variable():
    framework = koalak.mkframework()
    framework.create_variable("name", "test")
    assert framework.get_variable("name") == "test"


@pytest.mark.skip
def test_init_variable_in_mkframework():
    framework = koalak.mkframework(variables={"age": 30})
    assert framework.get_variable("age") == 30


def test_set_variable():
    framework = koalak.mkframework()
    framework.create_variable("name", "test")
    assert framework.get_variable("name") == "test"
    framework.set_variable("name", 30)
    assert framework.get_variable("name") == 30


# TODO: add list variables



import tempfile

import pytest
from koalak.databases import ListTxtDatabase


def test_simple():
    with tempfile.NamedTemporaryFile() as f:
        db = ListTxtDatabase(f.name)
        assert len(db) == 0
        assert db.count() == 0

        db.insert("a")
        assert len(db) == 1
        assert db.count() == 1
        assert db.list() == ["a"]

        db.insert("b")
        assert len(db) == 2
        assert db.count() == 2
        assert db.list() == ["a", "b"]

        db.remove("a")
        assert len(db) == 1
        assert db.count() == 1
        assert db.list() == ["b"]


def test_remove():
    with tempfile.NamedTemporaryFile() as f:
        db = ListTxtDatabase(f.name)
        db.insert("a")
        db.insert("b")
        db.insert("c")

        assert db.list() == ["a", "b", "c"]

        db.remove("c")
        assert db.list() == ["a", "b"]


def test_remove_not_found():
    with tempfile.NamedTemporaryFile() as f:
        db = ListTxtDatabase(f.name)
        db.insert("a")

        with pytest.raises(TypeError):
            db.remove("b")


def test_contain():
    with tempfile.NamedTemporaryFile() as f:
        db = ListTxtDatabase(f.name)
        assert "a" not in db
        assert "b" not in db

        db.insert("a")
        assert "a" in db

        db.insert("b")
        assert "a" in db
        assert "b" in db
        assert "c" not in db

        db.insert("c")
        assert "a" in db
        assert "b" in db
        assert "c" in db

        db.remove("a")
        assert "a" not in db


def test_unique():
    with tempfile.NamedTemporaryFile() as f:
        db = ListTxtDatabase(f.name, unique=True)
        db.insert("a")
        db.insert("b")
        db.insert("c")

        assert db.list() == ["a", "b", "c"]

        db.remove("c")
        assert db.list() == ["a", "b"]

        with pytest.raises(TypeError):
            db.insert("a")


def test_unique_function():
    with tempfile.NamedTemporaryFile() as f:
        db = ListTxtDatabase(f.name, unique=lambda x: x.split("=")[0])
        db.insert("a=1")
        db.insert("b=2")
        db.insert("c=3")

        assert db.list() == ["a=1", "b=2", "c=3"]

        with pytest.raises(TypeError):
            db.insert("a=5")



import argparse


class ArgparseSubcmdHelper:
    """argparse class Helper for subcommands
    To add a subcommand x implement the following methods
        parser_x(self, parser)
        run_x(self, args)
        description_x: str

    cls attributes
    prog:
    _parser_main: add to the main parser
    _run_main(self, args): hook after parsing args, handling main variable before dispatch to subcmd
    """

    prog = None
    groups = None

    def _parser_main(self, parser):
        pass

    def _run_main(self, args):
        pass

    def _init_help(self):
        """If groups exist change the current help"""
        if not self.groups:
            return
        SPACES = "    "

        # header prog name and description
        help = f"usage: {self.prog} [-h] <subcommand>\n\n"
        prog_description = getattr(self, "description", "")
        help += f"{prog_description}\n\n"

        # get the longest command name for pretty print with tabulations
        grouped_commands_name = [
            cmd for e in self.groups.values() for cmd in e["commands"]
        ]
        all_commands_name = [e[4:] for e in dir(self) if e.startswith("run_")]
        ungrouped_printed_commands = [
            e for e in all_commands_name if e not in grouped_commands_name
        ]
        # max command length for pretty print (aligned)
        max_command_length = len(max(grouped_commands_name, key=len))

        # FIXME: check that ungrouped do not exist
        # add ungrouped commands
        if ungrouped_printed_commands:
            self.groups["ungrouped"] = {
                "commands": ungrouped_printed_commands,
                "description": "",
            }

        for group_name, group in self.groups.items():
            commands = group["commands"]
            group_description = group["description"]
            help += f"{group_name}: {group_description}\n"
            for cmd in commands:
                # check if the command exist!
                if not hasattr(self, f"run_{cmd}"):
                    raise TypeError(
                        f"Command {cmd!r} can't be in a group because it don't exist (consider implementing 'run_{cmd}')"
                    )
                description_command = getattr(self, f"description_{cmd}", "")
                help += f"{SPACES}{cmd.ljust(max_command_length)}{SPACES} {description_command}\n"
            help += "\n"

        # change the help function
        self.main_parser.print_help = lambda file=None: print(help, file=file)
        # self.main_parser.print_usage = self.main_parser.print_help

    def __init__(self):
        # create main parser
        self.main_parser = argparse.ArgumentParser(self.prog)

        # hook to add arguments for the main program
        self._parser_main(self.main_parser)

        # init the subcommande
        self.subparsers = self.main_parser.add_subparsers(dest="subcommand")
        for subcommand_name in [
            e[4:]
            for e in dir(self)
            if e.startswith("run_") and callable(getattr(self, e))
        ]:

            # get help if exist
            description = getattr(self, "description_" + subcommand_name, None)
            subparser = self.subparsers.add_parser(
                subcommand_name, help=description, description=description
            )
            parser_name = f"parser_{subcommand_name}"

            # add parser_<cmd> if exist (the parser_<cmd> is not required only run_<cmd> is)
            if hasattr(self, parser_name):
                parser_func = getattr(self, parser_name)
                parser_func(subparser)

        # TODO: see the case if parser_ is present but not run_
        # change help if groups attribute is present
        self._init_help()

    def parse_args(self, args=None):
        return self.main_parser.parse_args(args)

    def main(self, args=None):
        parsed_args = self.main_parser.parse_args(args)
        self._run_main(parsed_args)
        subcommand_name = parsed_args.subcommand
        if subcommand_name is None:
            self.main_parser.print_help()
            return
        try:
            subcommand = getattr(self, f"run_{subcommand_name}")
            if not callable(subcommand):
                raise TypeError(f"{subcommand_name} attribute must be callable")

        except AttributeError:
            if hasattr(self, f"parser_{subcommand_name}"):
                print(f"Command {subcommand_name} not implemented")
                self.main_parser.print_usage()
                return
            else:
                # FIXME: dead code, argparse handle this case
                print(f"Command {subcommand_name} don't exist")
                self.main_parser.print_usage()
                return
        subcommand(parsed_args)
        return



import os
from typing import Any

"""
databases
- list
    - txt
    - json
    - csv
- key_value_db
    - json (dict)
- txt (list)
- json
- csv

existing light db (without port connection) ie using file
- sqlite
- text/csv/json
- tinydb
- dataset
- pickledb
- shelve (stdlib)
- dbm (key/value)
https://docs.python.org/3/library/persistence.html
"""


class Database:
    pass


class KeyValueDatabase(Database):
    def get(self, key):
        pass

    def set(self, key, value):
        pass


class ListDatabase(Database):
    def __len__(self) -> int:
        pass

    def insert(self, element: Any):
        pass

    def remove(self, element: Any):
        pass


class ListTxtDatabase(ListDatabase):
    def __init__(self, dbpath: str, unique=None):
        self.dbpath = dbpath
        self.unique = unique
        # create db if don't exist
        if not os.path.isfile(dbpath):
            with open(dbpath, "w"):
                pass

    def __len__(self):
        return self.count()

    def insert(self, line: str):
        if self.unique:
            if self.unique is True:
                if line in self:
                    raise TypeError(f"line {line!r} already exist")
            else:
                # unique if a function that return a string
                # check if the result of this function is unique in the db
                unique_line = self.unique(line)
                for db_line in self:
                    if self.unique(db_line) == unique_line:
                        raise TypeError(f"Identifier {unique_line!r} already exist")
        if "\n" in line:
            raise ValueError(f"line must not contain newline '\n' in txt database")
        with open(self.dbpath, "a") as f:
            f.write(line)
            f.write("\n")

    def remove(self, line_to_remove: str):
        # FIXME: can speed this by creating a tmp file
        # https://thispointer.com/python-how-to-delete-specific-lines-in-a-file-in-a-memory-efficient-way/

        # If self.unique is a function we remove by the id (result of unique)
        # and not by all the value
        if callable(self.unique):
            unique_callable = True
            line_to_remove = self.unique(line_to_remove)
        else:
            unique_callable = False
        new_lines = []
        removed = False
        with open(self.dbpath) as f:
            for line in f:
                compare_line = line[:-1]
                if unique_callable:
                    compare_line = self.unique(compare_line)
                if line_to_remove != compare_line:
                    new_lines.append(line)
                else:
                    # when we find the line to remove
                    # we add all other lines to the file
                    removed = True
                    for line in f:
                        new_lines.append(line)
                    break

        # if we removed nothing raise an exception
        if not removed:
            raise TypeError(f"line {line_to_remove!r} not found")

        # write new lines to the file
        with open(self.dbpath, "w") as f:
            f.writelines(new_lines)

    def count(self):
        return len(self.list())

    def list(self):
        with open(self.dbpath) as f:
            return [e[:-1] for e in f.readlines()]

    def pop(self):
        pass

    def __iter__(self):
        with open(self.dbpath) as f:
            for line in f:
                yield line[:-1]

    def __contains__(self, item):
        return self.contain(item)

    def contain(self, line: str):
        for db_line in self:
            if line == db_line:
                return True
        return False



import datetime
import random

# ==================== TEST BUDGET ============
import sys
import tempfile

import attr
import pytest
from sqlalchemy import (
    Boolean,
    Column,
    Date,
    DateTime,
    ForeignKey,
    Integer,
    String,
    create_engine,
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

echo = len(sys.argv) > 1
# print("echo", echo)
engine = create_engine("sqlite:///:memory:", echo=echo)
Session = sessionmaker(bind=engine)
Base = declarative_base()


@attr.s
class BudgetBase:
    name: str = attr.ib()
    initial_balance: int = attr.ib(default=0)
    description: str = attr.ib(default=True, kw_only=True)
    creation_date = attr.ib(factory=datetime.datetime.utcnow)
    loan: bool = attr.ib(default=True, kw_only=True)
    frm = attr.ib(default=None, kw_only=True)
    balance = attr.ib(init=False)
    refunded_balance = attr.ib(init=False)

    def __attrs_post_init__(self):
        print("AAAAAAAAAAAAAAAa")
        self.balance = self.initial_balance
        self.refunded_balance = self.initial_balance


old_init = BudgetBase.__attrs_post_init__


def new_init(self):
    print("BBBBBBBBBBBBBBBBB")
    old_init(self)
    print(self)
    session.add(self)
    session.commit()


BudgetBase.__attrs_post_init__ = new_init


class Budget(BudgetBase, Base):
    __tablename__ = "budgets"

    id = Column(Integer, primary_key=True)
    name = Column(String)
    description = Column(String)
    creation_date = Column(DateTime, default=datetime.datetime.utcnow, nullable=False)
    loan = Column(Boolean, default=True, nullable=False)
    id_from = Column(Integer, ForeignKey("budgets.id"), nullable=True)

    # balance
    initial_balance = Column(Integer, nullable=False)
    balance = Column(Integer, nullable=False)
    refunded_balance = Column(Integer, nullable=False)

    @property
    def transactions(self):
        return session.query(Transaction).filter_by(id_budget=self.id)


@attr.s
class TransactionBase:
    budget: Budget = attr.ib()
    value: int = attr.ib()
    description: str = attr.ib(default="")
    affected: Budget = attr.ib(default=None)
    loan: bool = attr.ib(default=False)

    def __attrs_post_init__(self):
        pass


old_init2 = TransactionBase.__attrs_post_init__


def new_init2(self):
    old_init2(self)
    session.add(self)
    session.commit()


TransactionBase.__attrs_post_init__ = new_init2


class Transaction(TransactionBase, Base):
    __tablename__ = "transactions"
    id = Column(Integer, primary_key=True)
    id_budget = Column(Integer, ForeignKey("budgets.id"), nullable=False)
    id_affected_budget = Column(Integer, ForeignKey("budgets.id"), nullable=True)
    time = Column(DateTime, default=datetime.datetime.utcnow, nullable=False)
    value = Column(Integer, nullable=False)
    description = Column(String, nullable=False)
    loan = Column(Boolean, default=True, nullable=False)

    # transactions
    @property
    def budget(self):
        return session.query(Budget).filter_by(id=self.id_budget).first()

    @budget.setter
    def budget(self, value):
        self.id_budget = value.id


Base.metadata.create_all(engine)
session = Session()

bnp = Budget("bnp")
print(bnp)

t1 = Transaction(bnp, 100, "phone")
print(t1)

# =================== END TEST ================



import tempfile
from typing import Dict, List

import koala
import pytest
from koala import RelationalDB, relationaldb


def test_relationaldb_simple():
    def db_factory(name):
        db = relationaldb(name)

        @db.table("persons")
        class Person:
            name: str = db.attribute()
            age: int = db.attribute()

        db.init()
        return db, Person

    with tempfile.NamedTemporaryFile() as f:
        db, Person = db_factory(f.name)
        # Add 3 persons to the db
        Person("Foo", 25)
        Person("Bar", 28)
        baz = Person("Baz", 20)
        baz.age = 22
        # FIXME: db.close()?
        db.close()
        # recreate the db to check persistance
        db, Person = db_factory(f.name)
        foo = db.query(Person).filter_by(name="Foo").first()
        assert foo.name == "Foo"
        assert foo.age == 25

        baz = db.query(Person).filter_by(name="Baz").first()
        assert baz.name == "Baz"
        assert baz.age == 22


def test_relationaldb_query():
    def db_factory(name):
        db = relationaldb(name)

        @db.table("Person")
        class Person:
            name: str = db.attribute()
            age: int = db.attribute()

        db.init()

        return db, Person

    with tempfile.NamedTemporaryFile() as f:
        db, Person = db_factory(f.name)

        foo = Person("foo", 20)
        assert foo.name == "foo"
        assert foo.age == 20
        bar = Person("bar", 21)
        baz = Person("baz", 22)

        assert len(list(db.query(Person))) == 3
        assert db.query(Person).filter_by(name="foo").first() is foo
        assert db.query(Person).filter_by(age=21).first() is bar

        assert len(list(db.query(Person).filter(Person.age >= 21))) == 2
        db.close()
        # REBUILD THE DB TO CHECK PERSISTENCE
        db, Person = db_factory(f.name)
        foo = db.first(Person, name="foo")
        bar = db.first(Person, name="bar")
        baz = db.first(Person, name="baz")

        assert foo.name == "foo"
        assert foo.age == 20
        assert bar.name == "bar"
        assert bar.age == 21
        assert baz.name == "baz"
        assert baz.age == 22


def test_relationaldb_first():
    # use sqlalchemy
    db = relationaldb()

    @db.table("persons")
    class Person:
        name: str = db.attribute()
        age: int = db.attribute()

    db.init()

    foo = Person("foo", 20)
    bar = Person("bar", 21)
    baz = Person("baz", 22)
    assert db.first(Person, name="foo") is foo
    assert db.first("persons", name="bar") is bar
    assert db.first("persons", name="baz") is baz


def test_one_to_many_relation_not_null_no_backref_with_ref_as_cls():
    def db_factory(name):
        db = relationaldb(name)

        @db.table("persons")
        class Person:
            name: str = db.attribute()
            age: int = db.attribute()
            # animals = db.attribute(type=db.List["Animal"], backref=True)

        @db.table("animals")
        class Animal:
            owner: Person = db.attribute()
            name: str = db.attribute()

        db.init()

        return db, Person, Animal

    with tempfile.NamedTemporaryFile() as f:
        db, Person, Animal = db_factory(f.name)
        with db:
            foo = Person("foo", 20)
            bar = Person("bar", 21)
            baz = Person("baz", 22)
            assert db.first(Person, name="foo") is foo
            assert db.first("persons", name="bar") is bar
            assert db.first("persons", name="baz") is baz

            afoo = Animal(foo, "afoo")
            assert afoo.owner is foo
            assert afoo.name == "afoo"

            afoo.owner = bar
            assert afoo.owner is bar
            assert afoo.name == "afoo"
            # import ipdb; ipdb.set_trace()
        # Rebuild to test persistence
        db, Person, Animal = db_factory(f.name)
        with db:
            afoo = db.first(Animal, name="afoo")
            bar = db.first(Person, name="bar")
            assert afoo.owner is bar

    # TODO fix all tests with factory


def test_one_to_many_relation_not_null_no_backref_with_ref_as_cls_inverse_order():
    def db_factory(name):
        db = relationaldb(name)

        @db.table("animals")
        class Animal:
            owner: "Person" = db.attribute()
            name: str = db.attribute()

        @db.table("persons")
        class Person:
            name: str = db.attribute()
            age: int = db.attribute()
            # animals = db.attribute(type=db.List["Animal"], backref=True)

        db.init()

        return db, Person, Animal

    with tempfile.NamedTemporaryFile() as f:
        db, Person, Animal = db_factory(f.name)
        with db:
            foo = Person("foo", 20)
            bar = Person("bar", 21)
            baz = Person("baz", 22)
            assert db.first(Person, name="foo") is foo
            assert db.first("persons", name="bar") is bar
            assert db.first("persons", name="baz") is baz

            afoo = Animal(foo, "afoo")
            assert afoo.owner is foo
            assert afoo.name == "afoo"

            afoo.owner = bar
            assert afoo.owner is bar
            assert afoo.name == "afoo"
            # import ipdb; ipdb.set_trace()
        # Rebuild to test persistence
        db, Person, Animal = db_factory(f.name)
        with db:
            afoo = db.first(Animal, name="afoo")
            bar = db.first(Person, name="bar")
            assert afoo.owner is bar


def test_one_to_many_relation_not_null_no_backref_with_ref_as_str():
    def db_factory(name):
        db = relationaldb(name)

        @db.table("persons")
        class Person:
            name: str = db.attribute()
            age: int = db.attribute()

        @db.table("animals")
        class Animal:
            owner = db.attribute(type="Person")
            name: str = db.attribute()

        db.init()
        return db, Person, Animal

    with tempfile.NamedTemporaryFile() as f:
        db, Person, Animal = db_factory(f.name)
        with db:
            foo = Person("foo", 20)
            bar = Person("bar", 21)
            baz = Person("baz", 22)
            assert db.first(Person, name="foo") is foo
            assert db.first("persons", name="bar") is bar
            assert db.first("persons", name="baz") is baz

            afoo = Animal(foo, "afoo")
            assert afoo.owner is foo
            assert afoo.name == "afoo"

            afoo.owner = bar
            assert afoo.owner is bar
            assert afoo.name == "afoo"
            # TODO: test that it's not nullable!

        db, Person, Animal = db_factory(f.name)
        with db:
            afoo = db.first(Animal, name="afoo")
            bar = db.first(Person, name="bar")
            assert afoo.owner is bar
            assert afoo.name == "afoo"


def test_one_to_many_relation_null_no_backref_with_ref_as_cls():
    def db_factory(name):
        db = relationaldb(name)

        @db.table("persons")
        class Person:
            name: str = db.attribute()
            age: int = db.attribute()

        @db.table("animals")
        class Animal:
            name: str = db.attribute()
            owner: Person = db.attribute(default=None)

        db.init()
        return db, Person, Animal

    with tempfile.NamedTemporaryFile() as f:
        db, Person, Animal = db_factory(f.name)
        with db:
            foo = Person("foo", 20)
            bar = Person("bar", 21)
            baz = Person("baz", 22)
            assert db.first(Person, name="foo") is foo
            assert db.first("persons", name="bar") is bar
            assert db.first("persons", name="baz") is baz

            afoo = Animal("afoo", foo)
            assert afoo.owner is foo
            assert afoo.name == "afoo"

            afoo.owner = bar
            assert afoo.owner is bar
            assert afoo.name == "afoo"

            abar = Animal("abar")
            assert abar.owner is None
            assert abar.name == "abar"

        db, Person, Animal = db_factory(f.name)
        with db:
            foo = db.first(Person, name="foo")
            bar = db.first(Person, name="bar")
            baz = db.first(Person, name="baz")
            afoo = db.first(Animal, name="afoo")
            abar = db.first(Animal, name="abar")

            assert afoo.owner is bar
            assert afoo.name == "afoo"

            assert abar.owner is None
            assert abar.name == "abar"


def test_error_referencing_non_existing_attribute_before_init():
    db = relationaldb()

    @db.table("persons")
    class Person:
        name: str = db.attribute()

    with pytest.raises(ValueError):

        @db.table("animals")
        class Animal:
            name: str = db.attribute()
            owner: Person = db.attribute(default=None, ref="age")


@pytest.mark.skip
def test_one_to_many_relation_null_backref_with_ref_as_cls():
    # TODO: implement, one to many, and many to one with different relations side
    def db_factory(name):
        db = relationaldb(name)

        @db.table("persons")
        class Person:
            name: str = db.attribute()
            age: int = db.attribute()
            animals: db.List["Animal"] = db.attribute(ref="owner", factory=list)

        @db.table("animals")
        class Animal:
            name: str = db.attribute()
            owner: Person = db.attribute(default=None)

        db.init()
        return db, Person, Animal

    with tempfile.NamedTemporaryFile() as f:
        db, Person, Animal = db_factory(f.name)

        foo = Person("foo", 20)
        bar = Person("bar", 21)
        baz = Person("baz", 22)

        assert not foo.animals
        assert db.first(Person, name="foo") is foo
        assert db.first("persons", name="bar") is bar
        assert db.first("persons", name="baz") is baz

        afoo = Animal("afoo", foo)
        db.commit()
        assert afoo.owner is foo
        assert afoo.name == "afoo"
        assert afoo in foo.animals

        afoo.owner = bar
        assert afoo.owner is bar
        assert afoo.name == "afoo"

        abar = Animal("abar")
        assert abar.owner is None
        assert abar.name == "abar"


def test_self_referencing_one_to_many_no_backref():
    db = relationaldb()

    @db.table("persons")
    class Person:
        name: str = db.attribute()
        age: int = db.attribute()
        best_friend = db.attribute(type="Person", default=None)

    db.init()

    # TODO: add same test with ref
    # TODO: add this test with many to many
    # TODO: add test that rise error if nullable is true when impossible, like in self referencing, nullable must be True (or set it to False automatticaly)
    foo = Person("foo", 21)
    assert foo.name == "foo"
    assert foo.age == 21
    assert foo.best_friend is None

    bar = Person("bar", 22)
    assert bar.name == "bar"
    assert bar.age == 22
    assert bar.best_friend is None

    foo.best_friend = bar
    assert foo.best_friend is bar
    assert bar.best_friend is None


def notest_complecated_relations():
    # FINAL TEST TO PASS?
    # TODO: activate me (rename notest to test)
    db = relationaldb()

    @db.table("persons")
    class Person:
        name: str = db.attribute()
        # persons (recursive)
        # a person to love

        # crush (one-to-many)
        crush = db.attribute(type="Person")
        my_lovers = db.attribute(type=db.List["Person"], relation=crush)

        # best friend (one-to-many)
        best_friend = db.attribute(type="Person")
        best_friend_of = db.attribute(type=db.List["Person"], relation=best_friend)

        # best animal (one-to-many)
        best_animal = db.attribute(type="Animal")

        # worst animal (one-to-many)
        worst_animal = db.attribute(type="Animal")

        # friends (many-to-many) (A can be a friend of B and B is not a friend of A, not transitive)
        friends = db.attribute(type=db.List["Person"])
        friends_of = ...
        exes = db.attribute(type=db.List["Person"])
        # animals
        prefered_animals = db.attribute(type=db.List["Animal"])
        owned_animals = db.attribute(type=db.List["Animal"])

        # test with hate and haters both ways
        hate = db.attribute(type="Person")
        haters = db.attribute(type=db.List["Person"], relation=hate)

        # or
        haters = db.attribute(type=db.List["Person"], many=False)
        hate = db.attribute(type="Person", relation=haters)
        # TODO: write example in koala/examples/relationaldb.py convert doc sqlalchemy to my syntax

    @db.table("animals")
    class Animal:
        name: str = db.attribute()
        crush: str = db.attribute(type="Animal")
        friends: str = db.attribute(type=db.List["Animal"])
        # persons
        prefered_persons: db.List[Person] = db.attribute()
        owners: db.List[Person] = db.attribute()
        best_owner: Person = db.attribute()
        worst_owner: Person = db.attribute()

        # backref
        persons_that_prefer_me: db.List[Person] = db.attribute(
            backref="prefered_animals"
        )


def notest_budget_example_app():
    db = relationaldb()

    @db.table("budgets")
    class Budget:
        name: str = db.attribute()
        initial_balance: int = db.attribute(default=0)
        description: str = db.attribute(default="")

        balance: int = db.attribute(init=False)

        transactions = db.attribute(type=db.List["Transaction"], relation="budget")

    @db.table("transactions")
    class Transaction:
        budget: Budget = db.attribute()
        value: int = db.attribute()
        description: str = db.attribute()

    @db.table("categories")
    class Categorie:
        name: str = db.attribute()
        description: str = db.attribute(default="")

    @db.table("tags")
    class Tag:
        name: str = db.attribute()
        description: str = db.attribute(default="")

    @db.table("animal")
    class Animal:
        name: str
        age: int


# TEST INTERNALS STATE


def test_building_attributes_one_simple_table():
    db = relationaldb()
    building_class = db._building_class

    @db.table("persons")
    class Person:
        name: str = db.attribute()
        age: int = db.attribute()

    assert len(building_class) == 1
    assert "Person" in building_class
    building_cls_person = building_class["Person"]
    assert building_cls_person.name == "Person"
    assert building_cls_person.table_name == "persons"
    assert building_cls_person.orm_cls is Person
    assert building_cls_person.exist
    assert len(building_cls_person.attributes) == 0


def test_building_attributes_one_two_table():
    db = relationaldb()
    building_class = db._building_class

    @db.table("persons")
    class Person:
        name: str = db.attribute()
        age: int = db.attribute()

    assert len(building_class) == 1
    assert "Person" in building_class
    building_cls_person = building_class["Person"]
    assert building_cls_person.name == "Person"
    assert building_cls_person.table_name == "persons"
    assert building_cls_person.orm_cls is Person
    assert building_cls_person.exist
    assert len(building_cls_person.attributes) == 0

    @db.table("animals")
    class Animal:
        name: str = db.attribute()

    assert len(building_class) == 2
    assert "Person" in building_class
    assert "Animal" in building_class

    building_cls_person = building_class["Person"]
    assert building_cls_person.name == "Person"
    assert building_cls_person.table_name == "persons"
    assert building_cls_person.orm_cls is Person
    assert building_cls_person.exist
    assert len(building_cls_person.attributes) == 0

    building_cls_animal = building_class["Animal"]
    assert building_cls_animal.name == "Animal"
    assert building_cls_animal.table_name == "animals"
    assert building_cls_animal.orm_cls is Animal
    assert building_cls_animal.exist
    assert len(building_cls_animal.attributes) == 0


def test_building_attributes_two_table_with_relation():
    db = relationaldb()
    building_class = db._building_class

    @db.table("persons")
    class Person:
        name: str = db.attribute()
        age: int = db.attribute()

    assert len(building_class) == 1
    assert "Person" in building_class
    building_cls_person = building_class["Person"]
    assert building_cls_person.name == "Person"
    assert building_cls_person.table_name == "persons"
    assert building_cls_person.orm_cls is Person
    assert building_cls_person.exist
    assert len(building_cls_person.attributes) == 0

    @db.table("animals")
    class Animal:
        name: str = db.attribute()
        owner: Person = db.attribute()

    assert len(building_class) == 2
    assert "Person" in building_class
    assert "Animal" in building_class

    building_cls_person = building_class["Person"]
    assert building_cls_person.name == "Person"
    assert building_cls_person.table_name == "persons"
    assert building_cls_person.orm_cls is Person
    assert building_cls_person.exist
    assert len(building_cls_person.attributes) == 0

    building_cls_animal = building_class["Animal"]
    assert building_cls_animal.name == "Animal"
    assert building_cls_animal.table_name == "animals"
    assert building_cls_animal.orm_cls is Animal
    assert building_cls_animal.exist
    assert len(building_cls_animal.attributes) == 1

    owner_attribute = building_cls_animal.attributes["owner"]
    assert owner_attribute.name == "owner"
    assert not owner_attribute.many
    assert owner_attribute.atomic_type_name == "Person"
    assert owner_attribute.referenced_by is None
    assert owner_attribute.exist is True


def test_building_attributes_one_table_with_relation():
    db = relationaldb()

    building_class = db._building_class

    @db.table("persons")
    class Person:
        name: str = db.attribute()
        age: int = db.attribute()
        animal: "Animal" = db.attribute()

    print("building_class", building_class)
    assert len(building_class) == 2
    assert "Person" in building_class
    building_cls_person = building_class["Person"]
    assert building_cls_person.name == "Person"
    assert building_cls_person.table_name == "persons"
    assert building_cls_person.orm_cls is Person
    assert building_cls_person.exist

    assert len(building_cls_person.attributes) == 1

    animal_attribute = building_cls_person.attributes["animal"]
    assert animal_attribute.name == "animal"
    assert animal_attribute.exist
    assert not animal_attribute.many
    assert animal_attribute.atomic_type_name == "Animal"
    assert animal_attribute.referenced_by is None

    building_cls_animal = building_class["Animal"]
    assert not building_cls_animal.exist


def notest_relationaldb_dict():
    db = relationaldb()

    @db.register
    class Person:
        name: str
        animal: "Animal" = db.attribute(ref="owner")

    @db.register
    class Animal:
        name: str
        owner: Person

    foo = Person("foo")
    bar = Person("bar")
    ploofy = Animal("ploofy")

    foo.animal = ploofy
    assert ploofy.owner is foo


class Person:
    def __init__(self, name: str, animal=None):
        self._animal = None

        self.name = name
        self.animal = animal

    @property
    def animal(self):
        return self._animal

    @animal.setter
    def animal(self, animal):
        if animal is None:
            if self._animal:
                self._animal._owner = None
            self._animal = animal

        if getattr(self, "_animal", False):
            self._animal._owner = None
        self._animal = animal
        animal._owner = self

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"


class Animal:
    def __init__(self, name, owner=None):
        self.name = name
        self.owner = owner

    @property
    def owner(self):
        return self._animal

    @owner.setter
    def owner(self, owner):
        if getattr(self, "_owner", False):
            self._owner._animal = None
        self._owner = owner
        owner._animal = self

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name})"
